#include "skein/orbit.h"

#include "astro/stateVectorIndices.hpp"
#include "astro/orbitalElementConversions.hpp"
#include "astro/twoBodyMethods.hpp"

#include <glm/gtc/matrix_transform.hpp>

Orbit::Orbit()
{
    _keplerianElements.resize(6);
}

// some values cause singularities when they are zero. so, make them not zero
void Orbit::makeSafe(double& value)
{
    value = value == 0.0 ? 0.0000001 : value;
}

double Orbit::getSemiMajorAxis() const
{
    return _keplerianElements[astro::semiMajorAxisIndex];
}
void Orbit::setSemiMajorAxis(double semiMajorAxis)
{
    _keplerianElements[astro::semiMajorAxisIndex] = semiMajorAxis;
}

double Orbit::getEccentricity() const
{
    return _keplerianElements[astro::eccentricityIndex];
}
void Orbit::setEccentricity(double eccentricity)
{
    _keplerianElements[astro::eccentricityIndex] = eccentricity;
}

double Orbit::getInclination() const
{
    return _keplerianElements[astro::inclinationIndex];
}
void Orbit::setInclination(double inclination)
{
    makeSafe(inclination);
    _keplerianElements[astro::inclinationIndex] = inclination;
}

double Orbit::getArgumentOfPeriapsis() const
{
    return _keplerianElements[astro::argumentOfPeriapsisIndex];
}
void Orbit::setArgumentOfPeriapsis(double argumentOfPeriapsis)
{
    makeSafe(argumentOfPeriapsis);
    _keplerianElements[astro::argumentOfPeriapsisIndex] = argumentOfPeriapsis;
}

double Orbit::getLongitudeOfAscendingNode() const
{
    return _keplerianElements[astro::longitudeOfAscendingNodeIndex];
}
void Orbit::setLongitudeOfAscendingNode(double longitudeOfAscendingNode)
{
    makeSafe(longitudeOfAscendingNode);
    _keplerianElements[astro::longitudeOfAscendingNodeIndex] = longitudeOfAscendingNode;
}

double Orbit::getTrueAnomaly() const
{
    return _keplerianElements[astro::trueAnomalyIndex];
}
void Orbit::setTrueAnomaly(double trueAnomaly)
{
    _keplerianElements[astro::trueAnomalyIndex] = trueAnomaly;
}

double Orbit::getMeanAnomalyAtEpoch() const
{
    return _meanAnomalyAtEpoch;
}
void Orbit::setMeanAnomalyAtEpoch(double meanAnomalyAtEpoch)
{
    _meanAnomalyAtEpoch = meanAnomalyAtEpoch;
}

const double Orbit::getEccentricAnomaly() const
{
    const double trueAnomaly = getTrueAnomaly();
    const double eccentricity = getEccentricity();
    double eccentricAnomaly = astro::convertTrueAnomalyToEllipticalEccentricAnomaly(trueAnomaly, eccentricity);

    return eccentricAnomaly;
}

glm::dvec3 Orbit::getPosition(double gravitationalParameter) const
{
    const Vector6 cartesian = getCartesianCoordinates(gravitationalParameter);
    return glm::dvec3(
        cartesian[astro::xPositionIndex],
        cartesian[astro::yPositionIndex],
        cartesian[astro::zPositionIndex]);
}

glm::dvec3 Orbit::getVelocityFromMeanAnomaly(const double meanAnomaly, double gravitationalParameter) const
{
    Vector6 cartesian = getCartesianCoordinates(gravitationalParameter);
    return glm::dvec3(
        cartesian[astro::xVelocityIndex],
        cartesian[astro::yVelocityIndex],
        cartesian[astro::zVelocityIndex]);
}

glm::dvec3 Orbit::getVelocity(double gravitationalParameter, double time) const
{
    return getVelocityFromMeanAnomaly(getMeanAnomaly(gravitationalParameter, time), gravitationalParameter);
}

const double Orbit::getMeanAnomaly(const double gravitationalParameter, const double time) const
{
    double meanMotion = astro::computeKeplerMeanMotion(getSemiMajorAxis(), gravitationalParameter);
    return meanMotion * time;
}

const Vector6 Orbit::getCartesianCoordinates(const double gravitationalParameter) const
{
    return astro::convertKeplerianToCartesianElements(_keplerianElements, gravitationalParameter);
}

void Orbit::update(double time, double gravitationalParameter)
{
    // compute mean anomaly from time and mean anomaly at epoch
    const double semiMajorAxis = getSemiMajorAxis();
    const double meanMotion = astro::computeKeplerMeanMotion(semiMajorAxis, gravitationalParameter);
    const double meanAnomaly = meanMotion * time + _meanAnomalyAtEpoch;

    // compute elliptical eccentric anomaly from mean anomaly
    const double eccentricity = getEccentricity();
    const double eccentricAnomaly = astro::convertEllipticalMeanAnomalyToEccentricAnomalyBS(
        eccentricity,
        meanAnomaly,
        200);

    setTrueAnomaly(astro::convertEccentricAnomalyToTrueAnomaly(eccentricAnomaly, eccentricity));
}

