#include "orbitvisualizer.hpp"
#include "gfx.hpp"

OrbitVisualizer::OrbitVisualizer(const ParticleMap& map, const std::string& particleId, const GLuint shaderProgram, float scale)
    : _map(map), _particleId(particleId), _shaderProgram(shaderProgram), _scale(scale)
{
    glGenVertexArrays(1, &_vao);
    glGenBuffers(1, &_vbo);
}

void OrbitVisualizer::render(const float time)
{
    // Orbit needs to be drawn relative to its parent
    const Particle& parent = _map.getParent(_particleId);
    const glm::vec3 parentPos = _map.getParticlePosition(parent.getId());

    regenerateVertices(parentPos);

    glUseProgram(_shaderProgram);
    updateModelViewProjectionMatrix(_shaderProgram, time);

    glBindVertexArray(_vao);
    glBindBuffer(GL_ARRAY_BUFFER, _vbo);

    glDrawArrays(GL_LINE_LOOP, 0, _vertices.size() / 3);
}

void OrbitVisualizer::regenerateVertices(const glm::vec3& basePos)
{
    Orbit orbit(_map.getOrbit(_particleId));

    _vertices.clear();
    for (int i = 0; i < _vertexCount; i++)
    {
        // TODO: this method of getting ellipse vertices is a huge hack. it would be
        // better to actually create a first-class ellipse object and use that to generate
        // a nice continuous mesh, instead of using orbital positions. 
        float t = (float)i / (float)_vertexCount * 2.0 * _pi;
        orbit.setTrueAnomaly(t);
        //glm::vec3 pos = orbit.getPositionFromMeanAnomaly(t, 1);
        glm::vec3 pos = orbit.getPosition(1);
        pos += basePos;

        // Vertices come out of the library with X and Y being in the 'flat' plane. Re-order them
        // here such that Z is up.
        float y = pos.z;
        pos.z = pos.y;
        pos.y = y;
        pos.z *= -1;

        pos /= _scale;

        _vertices.push_back(pos.x);
        _vertices.push_back(pos.y);
        _vertices.push_back(pos.z);
    }

    glBindVertexArray(_vao);

    glBindBuffer(GL_ARRAY_BUFFER, _vbo);
    size_t vboBufferSize = _vertices.size() * sizeof(float);
    glBufferData(GL_ARRAY_BUFFER, vboBufferSize, &_vertices[0], GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}

OrbitVisualizer::~OrbitVisualizer()
{
    glDeleteVertexArrays(1, &_vao);
    glDeleteBuffers(1, &_vbo);
}
